#!/usr/sbin/perl
#
# EventPair v1.0
# Last modified August 4, 2003 16:14
# (c) Olaf R.P. Bininda-Emonds
#
# Usage 'perl PairScript inFile charStates (odd; between 1 and 51)'
#
# Data contained in a tab-delimited text file in the form "SpeciesName Event1 ... Eventn"

#die "Must specify filename for input of event data and number of character states desired\n" if (@ARGV != 2);
#die "Number of character states must be odd and equal to or less than 51\n" if (@ARGV != 2);

#$inFile = $ARGV[0];
#$charStates = $ARGV[1];

$inFile = "MammalData.txt";
$charStates = 3;
#$coding = "event";

die "Number of character states must be odd\n" if (int($charStates/2)*2 == $charStates);
die "Number of character states must be between 1 and 51\n" if ($charStates < 1 || $charStates > 51);

open (fIN,"<$inFile");
open (fOUT,">$inFile.src");

while (<fIN>)	#Kludgy subroutine to parse datafile for number of taxa and events
	{
	$NumTaxa++ unless ($_ =~ /^\t/ | /^\n/);	#Ignore (any) completely blank lines
	if ($. == 1)
		{
		chop;
#		print "Last event for first taxon is whitespace. Check if really missing.\n" if ($_ =~/\s$/);
		while ($_ =~ /\t/g)
			{
			$NumEvent++;
			}
		}
	}
close fIN;	# To reset $.

$NumEP = $NumEvent*($NumEvent-1)/2;
for (1..$NumEP)
	{
	$AllEvents[$_] = $_;
	}
$range = $NumEvent / (($charStates - 1)/2);

$DelCount = 0;

print fOUT "#nexus\n\nbegin data;\n";
print fOUT "\tdimensions ntax = $NumTaxa nchar = $NumEP;\n";
for (1..($charStates-1)/2)
	{
	$PAUPsymbols1 .= chr(90-(($charStates-1)/2)+$_);
	$PAUPsymbols2 .= chr(96+$_);
	}
print fOUT "\tformat missing = ? respectcase symbols = \"1$PAUPsymbols1$PAUPsymbols2\";\n";
print fOUT "\n\tmatrix\n\n";

open (fIN,"<$inFile");
while (<fIN>)
	{
	unless ($_ =~ /^\t/ | /^\n/)	#Ignore (any) completely blank lines
		{
#		chop;
		$EPdata = "";
#		$CharNum = 1;
		($SpeciesName,@Events) = split(/\t/);
		$SpeciesName =~ s/\s+/_/g;	#Strip (any) white spaces out of taxon names
		print "$SpeciesName\n";
#		$NumEvent0 = $#Events+1;

		# Subroutine to convert data into ranks
			$LastEvent = 0;		#Kludgy subroutine to change missing data into last event (and therefore highest rank)
			for (0..$NumEvent-1)	#Get latest real value
				{
				$Events[$_] = "?" if ($Events[$_] eq "");	#Change missing data to ?
				$LastEvent = $Events[$_] if ($Events[$_] =~ /^-?\d+\.?\d*$/ and $Events[$_] > $LastEvent);		#Find highest event that is a decimal number
				}
			for (0..$NumEvent-1)
				{
				if ($Events[$_] =~ /^-?\d+\.?\d*$/)		#Test that data are decimal numbers
					{
					$datums[$_] = $Events[$_];
					}
				else
					{
					$datums[$_] = $LastEvent + 1;
					}
				}

			for (0..$NumEvent-1)	#Subroutine to generate index and rank tables
				{
				$preIT{$_+1} = $datums[$_];
				}
			@indexTable = sort {$preIT{$a} <=> $preIT{$b};} keys %preIT;
			$matchOne = 0;
			for (0..$NumEvent-1)
				{
				$SeqRankTable[$indexTable[$_]] = $_+1;
				$EventRankTable[$indexTable[$_]] = $_+1-$matchOne;	#Correction to give tied ranks the same rank without incrementing rank
				$matchOne++ if ($Events[$indexTable[$_]-1] == $Events[$indexTable[$_+1]-1]);
				}
			
			for (1..$NumEvent)	# Subroutine to correct for tied ranks
				{
				$rankList{$SeqRankTable[$_]} = $datums[$_-1];
				}
			$element = $NumEvent;
			while ($element > 1)
				{
				$match = 1;
				if ($rankList{$element} == $rankList{$element-1})
					{
					$topRank = $element;
					$RankSum = $element;
					until ($rankList{$element} != $rankList{$element-$match})
						{
						$RankSum += ($element-$match);
						$match++;
						}
					$TiedRank = $RankSum / $match;
					$fix = 1;

					for (1..$NumEvent)
						{
						for ($topRank-$match+1..$topRank)
							{
							if ($SeqRankTable[$fix] == $_)
								{
								$SeqRankTable[$fix] = $TiedRank;
								}
							}
						$fix++;
						}
					}
				$element -= $match;
				}

			for (0..$NumEvent-1)	#Changing ranks of missing data to missing
				{
				if ($datums[$_] == $LastEvent+1)
					{
					$SeqRankTable[$_+1] = ".";
					$EventRankTable[$_+1] = ".";
					}
				}
#			print "SeqRankTable\t@SeqRankTable\nEventRankTable\t@EventRankTable\nData\t\t@Events\n\n";
		# End of subroutine

		if ($coding eq "event")
			{
			@rankTable = @EventRankTable;
			}
		else
			{
			@rankTable = @SeqRankTable;
			}
		
		# Subroutine to event-pair data	
			$CharNum = 1;
			for ($RowEvent = 2; $RowEvent <= $NumEvent; $RowEvent++)
				{
				for ($ColEvent = 1; $ColEvent < $RowEvent; $ColEvent++)
					{
					if ($. == 1)
						{
						$CharSet[$RowEvent] .= $CharNum." ";		# Generate PAUP* charset
						$CharSet[$ColEvent] .= $CharNum." ";		# Generate PAUP* charset
						}
					$char = "";
					if ($rankTable[$RowEvent] =~ /^-?\d+\.?\d*$/ and $rankTable[$ColEvent] =~ /^-?\d+\.?\d*$/)		#Test that both ranks / events are decimal numbers
						{
						$ShiftMag = $rankTable[$RowEvent] - $rankTable[$ColEvent];
						if ($ShiftMag < 0)
							{
							$EPdata .= chr(int($ShiftMag/$range)+90);		#Assigns earlier events to upper case letters
							}
						elsif ($ShiftMag > 0)
							{
							$EPdata .= chr(int($ShiftMag/$range)+97);		#Assigns later events to lower case letters
							}
						else
							{
							$EPdata .= chr(49);		#The number 1
							}
						}
					else
						{
						$EPdata .= "?";
						}
					$CharNum++;
					if (int($CharNum / 100)*100 == $CharNum)	#Safety to control line length
						{
						$EPdata .= "\n\t";
						}
					}
				}
		# End of event-pair subroutine
						
		# Subroutine to find characters event is NOT involved in and produce short output
			if ($. == 1)
				{
				for ($Event = 1; $Event <= $NumEvent; $Event++)
					{
					(@InclSet) = split(/ /,$CharSet[$Event]);
					@InclSet = sort{$b <=> $a} @InclSet;
	
					@NotSet = @AllEvents;
					for (@InclSet)
						{
						splice(@NotSet,$_,1);
						}
					for (1..@NotSet)
						{
						$NotCharSet[$Event] .= $NotSet[$_]." ";		# Long version of not set
						}
	
					$count = 1;		# Short version of not set
					while ($count <= $#NotSet)
						{
						if ($NotSet[$count+1] - $NotSet[$count] == 1)
							{
							$lower = $count;
							while ($NotSet[$count+1] - $NotSet[$count] == 1)
								{
								$count++;
								}
							$ShortNotSet[$Event] .= $NotSet[$lower]."-";
							}
						else
							{
							$ShortNotSet[$Event] .= $NotSet[$count]." ";
							$count++;
							}
						}
					chop $ShortNotSet[$Event];
	
					$count = 0;		# Short version of charset
					(@LongCharSet) = split(/ /,$CharSet[$Event]);
					while ($count < $NumEvent)
						{
						if ($LongCharSet[$count+1] - $LongCharSet[$count] == 1)
							{
							$lower = $count;
							while ($LongCharSet[$count+1] - $LongCharSet[$count] == 1)
								{
								$count++;
								}
							$ShortCharSet[$Event] .= $LongCharSet[$lower]."-";
							}
						else
							{
							$ShortCharSet[$Event] .= $LongCharSet[$count]." ";
							$count++;
							}
						}
					chop $ShortCharSet[$Event];
					}
				}
		# End of not set subroutine

		print fOUT "$SpeciesName\t$EPdata\n";
		}
	}
close fIN;

print fOUT "\t;\nend;\n\n";

print fOUT "begin sets;\n";
for (1..$NumEvent)	#Print out charsets and charpartitions for each event
	{
	if ($_ == 1)
		{
		$SetName[$_] = $_."st";
		}
	elsif ($_ == 2)
		{
		$SetName[$_] = $_."nd";
		}
	elsif ($_ == 3)
		{
		$SetName[$_] = $_."rd";
		}
	else
		{
		$SetName[$_] = $_."th";
		}
	$Lower = ($_-1)*($NumEvent-1) + 1;
	$Upper = $_*($NumEvent-1);
	print fOUT "\tcharset $SetName[$_] = $ShortCharSet[$_];\n";
	print fOUT "\tcharpartition E$SetName[$_] = test:$SetName[$_], rest:$ShortNotSet[$_];\n";
	print fOUT "\n" unless ($_ == $NumEvent);
	}
print fOUT "end;\n";

print fOUT "\nbegin paup;\n";	#Print out search instruction block
print fOUT "\tlog file=EventPars.log replace;\n";
print fOUT "\tset autoclose=yes outroot=monophyl tcompress=yes notifybeep=no;\n";
print fOUT "\t[delete / only;]\n";		#In case any taxa are to be deleted
for (1..$NumEvent)
	{
	print fOUT "\n\t[!][!*** Event number: $_ ***]\n";
	print fOUT "\tinclude $SetName[$_] / only;\n";
#	print fOUT "\t\tbandb;\n";
	print fOUT "\t\thsearch / addseq=closest swap=nni;\n";
	print fOUT "\t\tcontree / strict=yes majrule=no adams=no semistrict=no indices=yes treefile=Events.contree ";
	if ($_ == 1)
		{
		print fOUT "replace;\n";
		}
	else
		{
		print fOUT "append;\n";
		}
	print fOUT "\t\tcleartrees nowarn=yes;\n\n";
	print fOUT "\t\tbootstrap nreps=1000 search=faststep keepall=yes grpfreq=no;\n";
	print fOUT "\t\tsavetrees file=Events.boottrees format=altnex savebootp=brlens maxdecimals=1 ";
	if ($_ == 1)
		{
		print fOUT "replace;\n";
		}
	else
		{
		print fOUT "append;\n";
		}
	print fOUT "\t\tcleartrees nowarn=yes;\n\n";
	print fOUT "\t\t[include all;\n";
	print fOUT "\t\thompart partition=E$SetName[$_] nreps=1000 / addseq=closest swap=nni;]\n";
	}

print fOUT "\n\t[!][*** All events included ***]\n";
print fOUT "\tinclude all;\n";
#print fOUT "\t\tbandb;\n";
print fOUT "\t\thsearch / addseq=closest swap=nni;\n";
print fOUT "\t\tsavetrees / format=altnexus file=FullEvent.tree replace;\n";
print fOUT "\t\tcontree / indices=yes treefile=FullEvent.contree replace;\n";
print fOUT "\n\tlog stop;\nend;\n";
#print fOUT "\nquit warntsave=no;\n";

close fOUT;

#system ("./paup $inFile.out > EventBoot.out");

open (fROUT,">rates.boot");
print fROUT "#nexus\nbegin trees;\n";
open (fRIN,"<Events.boottrees");	#Distill bootstrap tree descriptions and insert into rates blocks
while (<fRIN>)
	{
	if (m/tree PAUP_1 =/)
		{
		$TD=$_;
		print fROUT "$TD\n";
		}
	}
close fRIN;	
print fROUT "end;\n";
print fROUT "begin rates;\n";
print fROUT "\tsuper method=baum weights=yes;\nend;\n";
close fROUT;
		
#system("./r8s -b -f rates.boot > data.boot");

open (fBOOT, ">paup.boot");
	print fBOOT "begin paup;\n";
	print fBOOT "\tset autoclose=yes maxtrees=1000 increase=no notifybeep=no tcompress=yes torder=left outroot=monophyl criterion=parsimony status=no;\n";
	print fBOOT "\tlog file=Event.SuperBootTree;\n";

	open (fWTS, "<data.boot");	#Distilling bootstrap weightset from output
	while (<fWTS>)
		{
		if (m/:/)
			{
			$wts=$_;
			print fBOOT $wts;
			}
		}
	close fWTS;

	$BootSeed=int(10000*rand());

	print fBOOT "\thsearch addseq=random nreps=25 rseed=$BootSeed dstatus=no;\n";
	print fBOOT "\t\tcontree all / showtree=no treefile='Bootwt.tree' replace;\n";
	print fBOOT "\tlog stop;\n";
	print fBOOT "end;\n\nquit warntsave=no;\n";
close fBOOT;

#system ("cat data.boot paup.boot > paup.SuperBootSearch");
#system ("./paup paup.SuperBootSearch > SuperBoot.out");
